package de.cnam.jemina.database;

import de.cnam.jemina.config.SqlManager;
import de.cnam.jemina.objects.Artikel;
import de.cnam.jemina.persistents.IPersistent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.DriverManagerConnectionFactory;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.PoolingDriver;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.pool.KeyedObjectPoolFactory;
import org.apache.commons.pool.impl.GenericKeyedObjectPoolFactory;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Abstrake Klasse zum persistieren von Daten in eine Datenbank
 * @author CNAM Team (Nadine, Mike, Jens)
 */
public abstract class DatabasePersistents implements IPersistent {
        
    // Logging:
    private static final Logger log = LoggerFactory.getLogger(DatabasePersistents.class);
    
    // SQL Manager
    protected SqlManager sqlManager = SqlManager.getInstance();
    
    // Skripte
    private static final String SCRIPT_CREATE_TABLE = "createTabels.sql";
    
    // Tabellennamen:
    protected static final String TABLE_ARTIKEL = "artikel";
    protected static final String TABLE_BUCH = "buch";
    
    // Statements (aus der sqls.properties)
    protected static final String ARTIKEL_INSERT = "artikel_insert";
    protected static final String ARTIKEL_INSERT_BUCH = "artikel_insert_buch";
    protected static final String ARTIKEL_INSERT_ZEITSCHRIFT = "artikel_insert_zeitschrift";
    protected static final String ARTIKEL_INSERT_VIDEO = "artikel_insert_video";
    protected static final String ARTIKEL_INSERT_CD = "artikel_insert_cd";
    
    protected static final String ARTIKEL_SELECT_BYID = "artikel_select_byId";
    protected static final String ARTIKEL_SELECT_BYFILTER = "artikel_select_byFilter";
    protected static final String ARTIKEL_SELECT_BUCH_BYID = "artikel_select_buch_byId";
    protected static final String ARTIKEL_SELECT_CD_BYID = "artikel_select_cd_byId";
    protected static final String ARTIKEL_SELECT_VIDEO_BYID = "artikel_select_video_byId";
    protected static final String ARTIKEL_SELECT_ZEITSCHRIFT_BYID = "artikel_select_zeitschrift_byId";
    protected static final String ARTIKEL_SELECT_BYFILTER_WITH_OWNER = "artikel_select_byFilter_with_owner";
    protected static final String ARTIKEL_SELECT_ARTIKELID = "artikel_select_artikelId";
    
    protected static final String EXEMPLAR_CHANGE_ORT = "exemplar_change_ort";
    protected static final String EXEMPLAR_INSERT = "exemplar_insert";
    protected static final String EXEMPLAR_SELECT_BYARTIKELIDWITHOUTBESTAND = "exemplar_select_byArtikelIdWithoutBestand";
    protected static final String EXEMPLAR_SELECT_BYARTIKELID = "exemplar_select_byArtikelId";
    
    protected static final String BESITZER_SELECT = "besitzer_select";
    protected static final String BESITZER_SELECT_BYID = "besitzer_select_byId";
    protected static final String BESITZER_SELECT_FILIALE_BYID = "besitzer_select_filiale_byId";
    protected static final String BESITZER_SELECT_FILIALE = "besitzer_select_filiale";
    protected static final String BESITZER_SELECT_FILIALE_BYNAME = "besitzer_select_filiale_byName";
    protected static final String BESITZER_SELECT_PERSON_BYID = "besitzer_select_person_byId";
    protected static final String BESITZER_SELECT_PERSON_BYUSERID = "besitzer_select_person_byUserId";
    protected static final String BESITZER_SELECT_PERSON = "besitzer_select_person";
    protected static final String BESITZER_SELECT_MAXID = "besitzer_select_maxId";
    protected static final String BESITZER_INSERT = "besitzer_insert";
    protected static final String BESITZER_INSERT_FILIALE = "besitzer_insert_filiale";
    protected static final String BESITZER_INSERT_PERSON = "besitzer_insert_person";
    
    protected static final String RESERVIERUNG_STEP1 = "reservierung_step1";
    protected static final String RESERVIERUNG_STEP2 = "reservierung_step2";
    protected static final String RESERVIERUNG_INSERT = "reservierung_insert";
    protected static final String RESERVIERUNG_SELECT = "reservierung_select";
    protected static final String RESERVIERUNG_SELECT_BYNUMBER = "reservierung_select_byNumber";
    protected static final String RESERVIERUNG_SELECT_BYUSER = "reservierung_select_byUser";
    protected static final String RESERVIERUNG_DELETE_BYID = "reservierung_delete_byId";
    protected static final String RESERVIERUNG_DELETE_BYNUMBER = "reservierung_delete_byNumber";
    
    protected static final String AUSLEIHE_SELECT = "ausleihe_select";
    protected static final String AUSLEIHE_SELECT_BYSTATUS = "ausleihe_select_byStatus";
    protected static final String ausleihe_select_byUser = "ausleihe_select_byUser";
    protected static final String AUSLEIHE_SELECT_USERLIST = "ausleihe_select_userList";
    protected static final String AUSLEIHE_INSERT = "ausleihe_insert";
    protected static final String AUSLEIHE_UPDATE_STATUS = "ausleihe_update_Status";
    
    
    // Artikel-Typen:
    public static final String ARTIKEL_ZEITSCHRIFT = "Z";
    public static final String ARTIKEL_CD = "C";
    public static final String ARTIKEL_BUCH = "B";
    public static final String ARTIKEL_VIDEO = "V";
    
    // Besitzer Typen:
    public static final String BESITZER_FILIALE = "F";
    public static final String BESITZER_PERSON = "P";
    
    // Video Typen:
    public static final String VIDEO_TYPE_DVD = "D";
    public static final String VIDEO_TYPE_VHS = "V";
    public static final String VIDEO_TYPE_BLUERAY = "B";
    
    
    // JDBC URL
    private String jdbcUrl;
    
    /**
     * Liefert die maximale Zeichenl&auml;nge die in die Spalte passt
     * @param table Tabellenname
     * @param column Spaltenname
     * @return Liefert <b>0</b>, wenn keine Informationen vorliegen
     * @todo: Dies durch 'DatabaseMetaData' ermitteln lassen! Im Moment ist es einfach fest hinterlegt
     */
    protected int getColumnWidth(String table, String column) {
        if (table == null || column == null) {
            return 0;
        }
        
        // Artikel
        if (TABLE_ARTIKEL.equalsIgnoreCase(table)) {
            if ("TITEL".equalsIgnoreCase(column)) {
                return 255;
            }
            
            if ("BESCHREIBUNG".equalsIgnoreCase(column)) {
                return 5000;
            }
            
            if ("ERSTELLT_VON".equalsIgnoreCase(column)) {
                return 50;
            }
        }
        
        // Buch
        if (TABLE_BUCH.equalsIgnoreCase(table)) {
            if ("AUTOREN".equalsIgnoreCase(column)) {
                return 1000;
            }
            
            if ("ISBN".equalsIgnoreCase(column)) {
                return 50;
            }
        }
        
        return 0;
    }
    
    /**
     * Liefert zu dem Artikel die RowID.
     * Die Verbindung zu Datenbank wird <u>NICHT</u> geschlossen!
     * @param con Verbindung zur Datenbank
     * @param obj Zu suchendes Objekt
     * @return <b>-1</b> wenn dieser Artikel noch nicht in der Datenbank vorhanden ist
     */
    protected int getRowId(Connection con, Artikel obj) throws SQLException {
        int res = -1;
        
        PreparedStatement pre = con.prepareStatement(sqlManager.getSql(ARTIKEL_SELECT_ARTIKELID));
        pre.setString(1, StringUtils.abbreviate(obj.getTitel(), getColumnWidth(TABLE_ARTIKEL, "Titel")));
        try (ResultSet rs = pre.executeQuery()) {
            if (rs.next()) {
                res = rs.getInt(1);
            }
        }
        
        return res;
    }
    
    /**
     * Erstellt die Tabellen.
     * Dabei werden die Statements aus der Datei {@link #SCRIPT_CREATE_TABLE} ausgef&uuml;hrt
     * @param con Verbindung zur Datenbank
     * @todo: Pruefen ob Tabellen bestehen, im Moment versucht er sie immer anzulegen. Bei einer Exception hoert er auf
     */
    protected void createTabels(Connection con) {
        if (con == null) {
            log.warn("Keine Verbindung uebergeben");
            return;
        }
        try {
            BufferedReader in = new BufferedReader(new FileReader(new File(DatabasePersistents.class.getClassLoader().getResource(SCRIPT_CREATE_TABLE).toURI())));
            String zeile = null;
            while ((zeile = in.readLine()) != null) {
                if (! StringUtils.isEmpty(zeile)) {
                    if (log.isInfoEnabled()) {
                        log.info(zeile);
                    }
                    con.prepareStatement(zeile).execute();
                    con.commit();                    
                }
            }
            
            con.close();
        } catch (URISyntaxException | IOException | SQLException ex) {
            log.error("Fehler beim anlegen der Tabellen: {}", ex.getLocalizedMessage());
        }
    }
    
    
    /**
     * Inizialisiert den ConnectionPool
     * @param url JDBC URL
     * @param username Benutzername
     * @param password Passwort
     */
    protected void setupConnectionPool(String url, String username, String password) {
        try {
            GenericObjectPool gPool = new GenericObjectPool();
            ConnectionFactory connectionFactory = new DriverManagerConnectionFactory(url, username, password);
            KeyedObjectPoolFactory kopf = new GenericKeyedObjectPoolFactory(null, 10); // Max Verbindungen hier festlegen
            PoolableConnectionFactory pcf = new PoolableConnectionFactory(connectionFactory, gPool, kopf, null, false, true);

            // Verbindungen vorbelegen:
            for (int i=0;i<5;i++) {
                gPool.addObject();   
            }
            
            PoolingDriver pd = new PoolingDriver();
            final String poolName = "jemina";
            pd.registerPool(poolName, gPool);
            
            // URL fuer den Pool:
            jdbcUrl = "jdbc:apache:commons:dbcp:" + poolName;
            
        } catch (Exception ex) {
            log.error("Konnte DB-Pool nicht inizialisieren!", ex);
            jdbcUrl = url;
        }
    }

    
    /**
     * Liefert eine Verbindung zur Datenbank
     * @return Liefert <b>null</b> wenn keine Verbindung aufgebaut werden konnte
     * @throws SQLException 
     */
    @Override
    public Connection getConnection() throws SQLException {
        if (jdbcUrl == null) {
            log.error("Bitte erst den ConnectionPool inizialisieren!");
            return null;
        }
        return DriverManager.getConnection(jdbcUrl);
    }
    
}
