package de.cnam.jemina.persistents;

import de.cnam.jemina.config.SqlManager;
import de.cnam.jemina.database.DatabasePersistents;
import de.cnam.jemina.database.server.EmbeddedHsqlServer;
import de.cnam.jemina.extern.auth.IUser;
import de.cnam.jemina.objects.Artikel;
import de.cnam.jemina.objects.Buch;
import de.cnam.jemina.objects.Filiale;
import de.cnam.jemina.objects.Person;
import de.cnam.jemina.objects.Reservierung;
import de.cnam.jemina.objects.factory.ObjectFactory;
import de.cnam.jemina.stub.auth.User;
import java.sql.PreparedStatement;
import java.util.List;
import org.apache.commons.lang3.RandomStringUtils;
import org.junit.After;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 *
 * @author Administrator
 */
public class DatabaseTest {
    
    private static EmbeddedHsqlServer server; 
    private static IUser user;
    private static SqlManager sql;
    
    public DatabaseTest() {
    }
    
    @BeforeClass
    public static void setUpClass() {
        server = EmbeddedHsqlServer.getInstance();         
        user = new User("test", null, IUser.ROLLE.MITARBEITER, null);
        sql = SqlManager.getInstance();
    }
    
    @AfterClass
    public static void tearDownClass() {
        if (server != null) {
            server.stopDatabase();
        }
    }
    
    @Before
    public void setUp() {        
    }
    
    @After
    public void tearDown() {
        
    }
    
    /**
     * Testet ob die Datenbank korrekt hochgefahren wurde
     */
    @Test
    public void testServer() {
        assertNotNull(server);
        if (server != null) {
            assertNotNull(server.getConnection());
        }
    }
    
    /**
     * Testet ob ein Artikel korrekt eingef&uuml;gt wurde
     * @throws Exception 
     */
    @Test
    public void testArtikel() throws Exception {
        if (server != null) {
            // Create a dummy article
            Artikel a = createBook();
            
            // Insert
            server.getConnection().insert(a, user);
            
            // Select
            assertTrue( server.getConnection().selectArticle(a.getTitel()).size() > 0 );
            
        } else {
            // Server nicht korrekt gestartet
            throw new Exception("Server nicht korrekt gestartet");
        }
    }
    
    /**
     * Testet ob korrekt das Personen-Objekt erzeugt wird
     * @throws Exception 
     */
    @Test
    public void testPerson() throws Exception {
        if (server != null) {
            // Create a dummy user
            IUser testUser = createUser();
            
            // Select person (create a person object in database, if not exist)
            Person person = server.getConnection().getPerson(testUser);
            
            // Now select again
            Person person2 = server.getConnection().getPerson(testUser);
            
            // check equals
            assertEquals(person.getBesitzerId(), person2.getBesitzerId());
            
        } else {
            // Server nicht korrekt gestartet
            throw new Exception("Server nicht korrekt gestartet");
        }
    }
    
    /**
     * Testet die Reservierung
     * @throws Exception 
     */
    @Test
    public void testReserve() throws Exception {
        if (server != null) {
            /**
             * Test-User anlegen
             */
            IUser testUser = createUser();
            
            /**
             * Test-Artikel anlegen
             */
            Artikel a = insertArtikel();
            
            /**
             * Test-Filiale anlegen
             */
            Filiale f = insertFiliale();
            Filiale f2 = insertFiliale();
            
            /**
             * ### Reservierung Nr. 1 ###
             * Versuch zu Reservieren, ohne das es Exemplare gibt
             */
            assertNull(server.getConnection().reserve(a, f, testUser));
            
            /**
             * Lege ein Exemplar zu dem Artikel an.
             * Hierbei handelt es sich aber um ein Bestandsexemplar, welches nicht verliehen werden darf
             */
            insertExemplar(a, f, true);
            
            /**
             * ### Reservierung Nr. 2 ###
             * Versuch zu Reservieren, ohne das es ein freies Exemplare gibt.
             * Bestands-Exemplare duerfen nicht verliehen werden!
             */
            assertNull(server.getConnection().reserve(a, f, testUser));
            
                        
            /**
             * Lege ein Exemplar zu dem Artikel an (allerdings in einer anderen Filiale)
             */
            insertExemplar(a, f2, false);
            
            
            /**
             * ### Reservierung Nr. 3 ###
             * Versuch zu Reservieren, ohne das es ein freies Exemplare gibt.
             * Bestands-Exemplare duerfen nicht verliehen werden!
             * Nur Exemplare verleihbar, die auch in der Filiale sind!
             */
            assertNull(server.getConnection().reserve(a, f, testUser));
            
            
            /**
             * Lege ein Exemplar zu dem Artikel an, diesmal in der richtigen Filiale
             */
            insertExemplar(a, f, false);
            
            
            /**
             * ### Reservierung Nr. 4 ###
             * Reservierung muss nun erfolgreich sein
             */
            String rNumber = server.getConnection().reserve(a, f, testUser);
            assertNotNull(rNumber);
            
            
            
            /**
             * ### Reservierung Nr. 5 ###
             * Da das einzigste Exemplar reserviert ist, darf eine erneute Reservierung
             * nicht mehr funktionieren
             */
            assertNull(server.getConnection().reserve(a, f, testUser));
            
            
            /**
             * ### Liste Reservierung des Benutzers auf ###
             */
            List<Reservierung> rList = server.getConnection().getReserveFromUser(testUser);
            assertFalse( rList.isEmpty() );
            
            
            /**
             * Loesche Reservierung
             */
            server.getConnection().delReserve(rNumber);
            
            
            /**
             * ### Liste Reservierung des Benutzers auf ###
             * Diese sollten nun geringer sein
             */
            List<Reservierung> rList2 = server.getConnection().getReserveFromUser(testUser);
            assertTrue( rList2.size() <= rList.size() );
            
            
            /**
             * ### Reservierung Nr. 6 ###
             * Reservierung sollte wieder moeglich sein, da die alte Reservierung geloescht wurde
             */
            rNumber = server.getConnection().reserve(a, f, testUser);
            assertNotNull(rNumber);            
        } else {
            // Server nicht korrekt gestartet
            throw new Exception("Server nicht korrekt gestartet");
        }
    }
    
    /**
     * Testet das Ausleihen
     * @throws Exception 
     */
    @Test
    public void testLoan() throws Exception {
        if (server != null) {
            /**
             * Test-User anlegen
             */
            IUser testUser = createUser();
            
            /**
             * Test-Artikel anlegen
             */
            Artikel a = insertArtikel();
            
            /**
             * Test-Filiale anlegen
             */
            Filiale f = insertFiliale();
            
            /**
             * Lege ein Exemplar zu dem Artikel an.
             */
            insertExemplar(a, f, false);
            
            /**
             * Reservierung
             */
            Reservierung r = insertReservierung(a, f, testUser);
            
            /**
             * Lade Exemplar-Liste
             */
            List<Integer> exList = server.getConnection().getExemplarIds(a, false);
            assertFalse( exList.isEmpty() );
            
            /**
             * Ausleihe
             */
            server.getConnection().loan(r, exList.get(0));
            
            /**
             * Reservierung sollte nun nicht mehr existieren
             */
            assertNull( server.getConnection().getReserve(r.getNummer()) );            
        } else {
            // Server nicht korrekt gestartet
            throw new Exception("Server nicht korrekt gestartet");
        }
    }
    
    /**
     * Dieser Test versucht absichtlich Daten einzufuegen, die viel zu gross sind
     */
    @Test
    public void testMaxLength() throws Exception {
        /**
         * Artikel anlegen
         */        
        Artikel a = createBook();
        
        /**
         * Grosse String zuweisen
         */
        final String longTxt = RandomStringUtils.randomAlphanumeric(6000);
        a.setBeschreibung(longTxt);
        a.setTitel(longTxt);
        ((Buch)a).setAutoren(longTxt);
        ((Buch)a).setIsbn(longTxt);
        
        
        /**
         * Artikel einfuegen
         */
        server.getConnection().insert(a, user);
    }
    
    
    
    /**
     * F&uuml;hrt eine Reservierung durch
     * @param a Artikel
     * @param f Filiale
     * @param u User
     * @return
     * @throws Exception 
     */
    private Reservierung insertReservierung(Artikel a, Filiale f, IUser u) throws Exception {
        String rNumber = server.getConnection().reserve(a, f, u);
        
        return server.getConnection().getReserve(rNumber);
    }
    
    /**
     * Erstellt einen {@link IUser User} mit der Rolle Student und
     * einem zuf&auml;lligen Namen
     * @return 
     */
    private IUser createUser() {
        return new User(RandomStringUtils.randomAlphabetic(20), null, IUser.ROLLE.STUDENT, null);
    }
    
    /**
     * Erstellt ein Buch-Objekt mit zuf&auml;lligen Werten
     * @return 
     */
    private Artikel createBook() {
        Artikel a = ObjectFactory.genArtikel(-1, RandomStringUtils.randomAlphabetic(255), RandomStringUtils.randomAlphabetic(100), DatabasePersistents.ARTIKEL_BUCH);
        ((Buch) a).setAutoren(RandomStringUtils.randomAlphabetic(50));
        ((Buch) a).setIsbn(RandomStringUtils.randomNumeric(10));
        
        return a;
    }
    
    /**
     * F&uuml;gt einen Artikel in die Datenbank ein
     * @return 
     */
    private Artikel insertArtikel() throws Exception {
        Artikel a = createBook();
        /**
         * Artikel einfuegen
         */
        server.getConnection().insert(a, user);

        /**
         * Artikel selektieren (dadurch bekommen wir die Artikel-ID)
         */
        List<Artikel> aList = server.getConnection().selectArticle(a.getTitel());
        a = aList.get(0);
        
        return a;
    }
    
    /**
     * Erstellt ein Filiale-Objekt mit zuf&auml;lligen Werten
     * @return 
     */
    private Filiale createFiliale() {
        return new Filiale(-1, RandomStringUtils.randomAlphabetic(255), 0, RandomStringUtils.randomAlphabetic(10), RandomStringUtils.randomAlphabetic(10));
    }
    
    /**
     * F&uuml;gt ein Filiale in die Datenbank ein
     * @return 
     */
    private Filiale insertFiliale() throws Exception {
        Filiale f = createFiliale();
        
        /**
         * Filiale einfuegen
         */
        server.getConnection().insert(f, user);

        /**
         * Filialie selektieren (dadurch bekommen wir die Besitzer-ID)
         */
        f = server.getConnection().getFilialeByName(f.getName());
        
        return f;
    }
    
    
    /**
     * Legt ein Exemplar an
     * @param a Artikel zu dem ein Exemplar angelegt werden soll
     * @param f Filiale in der das Exemplar steht
     * @param bestand Bestands-Exemplar?
     * @throws Exception 
     */
    private void insertExemplar(Artikel a, Filiale f, boolean bestand) throws Exception {
        PreparedStatement pre = server.getConnection().getConnection().prepareStatement(sql.getSql("exemplar_insert"));
        pre.setInt(1, a.getId());
        pre.setBoolean(2, bestand);
        pre.setInt(3, f.getBesitzerId());
        pre.execute();
    }
    
            
}
